# OneTrigger - Trigger webhooks by Onedata events
# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os, logging, sys, time, datetime
import requests
from tabulate import tabulate


class OneproviderClient:

    CONNECTION_RETRIES = 3
    TIMEOUT = 10
    WAIT_SECONDS = 10
    SPACES_PATH = '/api/v3/oneprovider/spaces/'
    FILES_PATH = '/api/v3/oneprovider/files/'
    ATTRIBUTES_PATH = '/api/v3/oneprovider/attributes/'

    def __init__(self, settings):
        self._settings = settings
        self._header = {'X-Auth-Token': settings.token}

    def _check_space(self):
        self._settings.space = self._settings.space.strip('/')
        space_url = 'https://{0}{1}'.format(self._settings.host, self.SPACES_PATH)
        try:
            r = requests.get(space_url, headers=self._header, verify=not self._settings.insecure)
            if r.status_code == 200:
                spaces = r.json()
                space_names = []
                for space in spaces:
                    space_names.append(space['name'])
                if self._settings.space not in space_names:
                    logging.error('The space "{0}" does not exist. Please check the available spaces with the "list-spaces" subcommand.'.format(self._settings.space))
                    sys.exit(1)
            elif r.status_code == 401:
                logging.error('Invalid token')
                sys.exit(1)
            else:
                raise Exception
        except Exception:
            logging.error('Error connecting to provider host')
            sys.exit(1)

    def _check_folder(self):
        folder = None
        if self._settings.folder != None:
            folder = self._settings.folder.strip('/')
            folder_url = 'https://{0}{1}{2}/{3}'.format(self._settings.host, self.FILES_PATH, self._settings.space, folder)
            try:
                r = requests.get(folder_url, headers=self._header, verify=not self._settings.insecure)
                if r.status_code == 404:
                    logging.warning('The folder "{0}" does not exist. Listening events on space folder'.format(folder))
                    folder = None
                elif r.status_code != 200:
                    raise Exception
            except Exception:
                logging.error('Error connecting to provider host')
                sys.exit(1)
        return folder

    def _post_event(self, file_id, file_path):
        event = {
            'Key': file_path,
            'Records': [
                {
                    'objectKey': os.path.basename(file_path),
                    'objectId': file_id,
                    'eventTime': datetime.datetime.utcnow().isoformat(),
                    'eventSource': 'OneTrigger'
                }
            ]
        }
        try:
            r = requests.post(self._settings.webhook, json=event)
            logging.info('File "{0}" uploaded. Event sent to {1} - {2}'.format(
                event['file'], self._settings.webhook, r.status_code))
        except Exception:
            logging.error(
                'Error sending event to {0}'.format(self._settings.webhook))

    def _subscribe(self, folder):
        folder_msg = '' if folder == None else '. Listening events on "/{0}/{1}/" folder'.format(self._settings.space, folder)
        files = []
        initialized = False
        attempts = 0
        logging.info('Subscribing to file events in space "{0}" from provider "{1}"...{2}'.format(self._settings.space, self._settings.host, folder_msg))
        while attempts < self.CONNECTION_RETRIES:
            try:
                while True:
                    # Get all files in folder and subfolders
                    new_files = []
                    paths = ['{0}/{1}'.format(self._settings.space, folder)] if folder != None else [self._settings.space]
                    for path in paths:
                        files_url = 'https://{0}{1}{2}'.format(self._settings.host, self.FILES_PATH, path)
                        r = requests.get(files_url, headers=self._header, verify=not self._settings.insecure)
                        if r.status_code == 200:
                            path_items = r.json()
                            for item in path_items:
                                if item['id'] in files:
                                    new_files.append(item['id'])
                                    continue
                                attributes_url = 'https://{0}{1}{2}'.format(self._settings.host, self.ATTRIBUTES_PATH, item['path'].lstrip('/'))
                                r = requests.get(attributes_url, headers=self._header, verify=not self._settings.insecure)
                                if r.status_code == 200:
                                    attributes = r.json()
                                    # Check subfolders
                                    if attributes['type'].lower() == 'dir':
                                        paths.append(item['path'])
                                    if attributes['type'].lower() == 'reg':
                                        new_files.append(item['id'])
                                        if initialized and item['id'] not in files:
                                            self._post_event(item['id'], item['path'])
                                else:
                                    raise Exception  
                        else:
                            raise Exception
                    attempts = 0
                    files = new_files
                    #logging.info('Connected - {0} files'.format(len(files)))
                    initialized = True
                    time.sleep(self.TIMEOUT)
            except Exception:
                logging.warning('Connection lost. Retrying...')
            attempts += 1
            time.sleep(self.WAIT_SECONDS)
        logging.error('Connection error')
        sys.exit(1)

    def run(self):
        self._check_space()
        folder = self._check_folder()
        self._subscribe(folder)

    def _print_spaces(self, spaces):
        table = []
        for space in spaces:
            table.append([space['name'], space['spaceId']])
        print(tabulate(table, headers=['Name', 'Space ID']))

    def list_spaces(self):
        list_spaces_url = 'https://{0}{1}'.format(self._settings.host, self.SPACES_PATH)
        try:
            r = requests.get(list_spaces_url, headers=self._header, verify=not self._settings.insecure)
            if r.status_code == 200:
                spaces = r.json()
                self._print_spaces(spaces)
            elif r.status_code == 401:
                logging.error('Invalid token')
                sys.exit(1)
            else:
                raise Exception
        except Exception:
            logging.error('Error connecting to provider host')
            sys.exit(1)



# OneTrigger - Trigger webhooks by Onedata events
# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os, json, argparse, logging, sys
from onetrigger.version import __version__


class ConfigParser:

    def __init__(self):
        self._parse_env_config()
        self._create_parser()
        self._create_subparsers()

    def _parse_env_config(self):
        self._env_host = os.environ.get('ONEPROVIDER_HOST')
        self._env_token = os.environ.get('ONEDATA_ACCESS_TOKEN')
        self._env_space = os.environ.get('ONEDATA_SPACE')
        self._env_webhook = os.environ.get('ONETRIGGER_WEBHOOK')
        self._env_folder = os.environ.get('ONEDATA_SPACE_FOLDER')
        self._env_insecure = json.loads(os.environ.get('ONEPROVIDER_INSECURE').lower()) if 'ONEPROVIDER_INSECURE' in os.environ else False

    def _create_parser(self):
        self._parser = argparse.ArgumentParser(description='Trigger webhooks by Onedata events')
        self._parser.add_argument('-v', '--version', action='version', version='version {0}'.format(__version__),
                            help='show OneTrigger version')

        # Parent parser to describe shared arguments
        self._parent_parser = argparse.ArgumentParser(add_help=False)
        self._parent_parser.add_argument('-H', '--oneprovider-host', action='store', default=self._env_host, dest='host',
                            help='Oneprovider hostname or IP')
        self._parent_parser.add_argument('-t', '--token', action='store', default=self._env_token, dest='token',
                            help='Onedata access token')
        self._parent_parser.add_argument('-i', '--insecure', action='store_true', default=self._env_insecure, dest='insecure',
                            help='Connect to a provider without a trusted certificate (Optional)')

    def _create_subparsers(self):
        self._subparsers = self._parser.add_subparsers(title='Commands', dest='command')
        self._subparsers.required = True
        self._create_run_parser()
        self._create_list_spaces_parser()

    def _create_run_parser(self):
        self._run_parser = self._subparsers.add_parser('run', parents=[self._parent_parser], help='Run OneTrigger')
        self._run_parser.add_argument('-s', '--space', action='store', default=self._env_space, dest='space',
                            help='Onedata space')
        self._run_parser.add_argument('-w', '--webhook', action='store', default=self._env_webhook, dest='webhook',
                            help='Webhook to send events')
        self._run_parser.add_argument('-f', '--folder', action='store', default=self._env_folder, dest='folder',
                            help='Folder to listen events (Optional)')

    def _create_list_spaces_parser(self):
        self._list_spaces_parser = self._subparsers.add_parser('list-spaces', parents=[self._parent_parser], help='List available spaces')

    def _check_config(self, config):
        required = True
        if config.host == None:
            logging.error('Oneprovider host is not provided. Please set it via "--oneprovider-host" argument or "ONEPROVIDER_HOST" environment variable')
            required = False
        if config.token == None:
            logging.error('Onedata access token is not provided. Please set it via "--token" argument or "ONEDATA_ACCESS_TOKEN" environment variable')
            required = False
        if config.command == 'run':
            if config.space == None:
                logging.error('Onedata space is not provided. Please set it via "--space" argument or "ONEDATA_SPACE" environment variable')
                required = False
            if config.webhook == None:
                logging.error('Webhook to send events is not provided. Please set it via "--webhook" argument or "ONETRIGGER_WEBHOOK" environment variable')
                required = False
        if required == False:
            sys.exit(1)

    def parse(self):
        config = self._parser.parse_args()
        self._check_config(config)
        return config




# OneTrigger - Trigger webhooks by Onedata events
# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
 
import logging, sys, signal
from onetrigger.configparser import ConfigParser
from onetrigger.oneproviderclient import OneproviderClient


class OneTrigger:

    def __init__(self):
        self._config_logging()
        self._set_sig_handlers()
        self._config = ConfigParser().parse()
        self._oneprovider_client = OneproviderClient(self._config)

    def _config_logging(self):
        LOGGING_FORMAT = '%(asctime)s - %(levelname)s - %(message)s'
        logging.basicConfig(format=LOGGING_FORMAT, level=logging.INFO)

    def _set_sig_handlers(self):
        def sig_handler(sig, frame):
            logging.info('Closing OneTrigger... Bye!')
            sys.exit(0)
        for sig in (signal.SIGINT, signal.SIGTERM):
            signal.signal(sig, sig_handler)

    def _run(self):
        self._oneprovider_client.run()

    def _list_spaces(self):
        self._oneprovider_client.list_spaces()

    def main(self):
        {
            'run': self._run,
            'list-spaces': self._list_spaces
        }[self._config.command]()

def main():
    OneTrigger().main()

if __name__ == '__main__':
    main()